(******************************************************************************)
(** * SRAG simulates SRAL: read step   *)
(******************************************************************************)

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Require Import Basic.
Require Import Labels Events Execution.
Require Import SRALn Wmax SRAG.
Require Import Write_lists Match_SRAn.

Require Import FG.

Set Implicit Arguments.

Lemma GsimL_read_helper_p G (WF_G : Wf G)
 n G' t e (ADD : add G G' t n (Aload (loc e) (valw e)) e)
 ol wl p (NEQ: p <> t) (MATCH : SGmatch_wl G p ol wl) :
 SGmatch_wl G' p ol wl.
Proof.
red in MATCH; desc.
red in WL_WF; desc.
red; splits.
- red; splits; try done.
  eby rewrite WL_G; eapply add_acts_monotone.
  setoid_rewrite add_non_write_co with (G:=G); try edone.
  type_solver.
- red in WL_CONS; desc; red; splits.
  + rewrite add_non_write_dom_co_hb; try edone.
    2: type_solver.
    revert WL_D; unfold is_tid; basic_solver 12.
  + ins. rewrite add_non_write_co_hb; try edone.
    2: type_solver.
    relsf; apply irreflexive_union; splits; try done.
    rewrite (dom_l (event_order_wlE _ _)), WL_G.
    unfolder; ins; apply (add_fresh ADD); desf.
Qed.

Lemma GsimL_read_helper_t G (WF_G : Wf G)
 n G' t u e x v tw
 (LOC: loc e = x) (VAL: valw e = v) (TID: tid e = tw)
 (ADD : add G G' t n (Aload x v) e)
 ol (Z : 0 < length ol) wl
 (MATCH : SGmatch_wl G t ol wl)
 (WL_RMW : u = true -> ~ dom_rel (co G) e)
 (MIN: (hb G)^? e (wl 0 x)) :
 SGmatch_wl G' t ol wl.
Proof.
cdes MATCH; cdes WL_WF.
assert (WL_TID' := WL_TID).
specialize (WL_TID 0 Z); ins.
assert (WL_LOC' := WL_LOC).
specialize (WL_LOC 0 Z); ins.
assert (WL_VAL' := WL_VAL).
specialize (WL_VAL 0 Z); ins.
assert (WL_RMW' := WL_RMW0).
specialize (WL_RMW0 0 Z); ins.
ins; desf.
red; splits.
- red; splits; eauto.
  + eby rewrite <- add_acts_monotone.
  + setoid_rewrite add_non_write_co with (G:=G); try edone.
    type_solver.
- cdes WL_CONS.
  red; splits.
  + rewrite add_non_write_dom_co_hb; try edone.
    2: type_solver.
    relsf; splits; [by auto|].
    arewrite (dom_rel (co G  ;; (hb G)^? ⨾ ⦗eq e⦘) ⊆₁ 
        dom_rel (co G  ;; (hb G)^? ⨾ ⦗eq (wl 0 (loc e))⦘)).
    { revert MIN; generalize (@hb_trans G). 
      clear. basic_solver 12. }
    simpl.
    remember (S (length ol)) as N.
    rewrite <- seqA.
    remember (co G ⨾ (hb G)^?) as D.
    unfolder; ins; desc; subst y. rename x into e'.
    eapply (WL_O (wl 0 (loc e))).
    exists e'; splits; auto.
    eapply event_order_zero; auto.
  + rewrite add_non_write_co_hb; try edone.
    2: type_solver.
    relsf; splits.
    apply irreflexive_union; splits; [done|].
    rewrite (dom_l (event_order_wlE _ _)).
    rewrite WL_G.
    unfolder; ins; apply (add_fresh ADD); desf.
Qed.

Lemma GsimL_read (st st' : State) 
          (G : execution) (WF_G: Wf G)
          t x v (SRAL: SRAL_step st st' (Some (t, (Aload x v))))
          (MATCH: SGmatch st G) : 
          exists G', SRAG_step G G' (Some (t, (Aload x v))) /\ 
                     SGmatch st' G'.
Proof.
inversion SRAL; subst; clear SRAL.
red in READ; subst; desf.
forward apply SGmatch_SGmatch_fun; try edone; ins; desc.
rename H into FWL.
assert (FWL_COPY := FWL).
rename st' into st.
assert (WF_S : wf_state st) by eapply SGmatch_implies_wf_st, MATCH.
assert (exists w ol_min,
  << IN_MIN: In ol_min (st t) >> /\
  << W_FIRST: fwl t ol_min 0 x = w >> /\
  << MIN : forall ol (IN: In ol (st t)), (hb G)^? w (fwl t ol 0 x)>>).
{ red in WF_S; desc.
  assert (IN: exists ol, In ol (st t)).
  by specialize (NON_EMPTY_POT t); destruct (st t); ins; eauto.
  desc.
  destruct ol as [|d L]; desf.
  by exfalso; eapply NON_EMPTY_OL; eauto; red; vauto.
  red in MATCH; desc.
  set (EE := fun e => exists ol, In ol (st t) /\ fwl t ol 0 x = e).
  assert (INe: In (fwl t (d :: L) 0 x) G.(acts)).
  { eapply FWL; eauto. apply event_in_wl_k. ins; lia. }
  set (s := (sb G)^{-1} ;; ⦗EE⦘).
  assert (MIN: exists e0, s^* (fwl t (d :: L) 0 x) e0 /\ max_elt s e0).
  { apply last_exists with (dom := G.(acts)).
    - unfold s.
      arewrite ((sb G)⁻¹ ⨾ ⦗EE⦘ ⊆ (sb G)⁻¹).
      red.
      rewrite <- transp_ct.
      rewrite ct_of_trans.
      unfolder; ins.
      eby eapply sb_irr.
      apply sb_trans.
    - ins.
      apply rt_of_trans in H.
      unfold s in H.
      hahn_rewrite wf_sbE in H.
      revert H; basic_solver.
      unfold s.
      red; unfolder; ins; desf.
      exists z; splits; eauto.
      eapply sb_trans; eauto. }
  assert (SM:= SMATCH t (d :: L) IN).
  destruct MIN as [e0 [S MAX]].
  unfold s in *; clear s.
  assert (EE0: EE e0).
  { apply rtE in S.
    hahn_rewrite inclusion_ct_seq_eqv_r in S.
    unfolder in S; desf.
  unfold EE; eexists; splits; eauto. }
  clear S.
  rename e0 into e.
  unfold EE in *; clear EE.
  destruct EE0 as [ol' [IN0 F1]].
  exists e, ol'; splits; try done.
  ins.
  clear d IN INe SM L.
  destruct ol as [|d L].
  by exfalso; eapply NON_EMPTY_OL; eauto; red; vauto.
  destruct ol' as [|d' L'].
  by exfalso; eapply NON_EMPTY_OL with (L:=nil); eauto.
  subst.
  assert (FWL' := FWL).
  specialize (FWL _ _ IN1).
  specialize (FWL' _ _ IN0).
  desc.
  assert (A: (⦗acts_set G⦘ ⨾ same_tid ⨾ ⦗acts_set G⦘) 
    (fwl t (d' :: L') 0 x) (fwl t (d :: L) 0 x)).
  { clear MAX. 
    unfolder; ins; splits.
    + red in FWL'; desc.
      apply WL_G.
      by eapply event_in_wl_k; ins; lia.
    + red.
      red in FWL; desc.
      rewrite WL_TID; [|ins;lia].
      red in FWL'; desc.
      rewrite WL_TID0; [|ins;lia].
      simpl.
      replace (d x) with (RT tw v u) by (symmetry; eauto).
      by replace (d' x) with (RT tw v u) by (symmetry; eauto).
    + red in FWL; desc.
      apply WL_G.
      by eapply event_in_wl_k; ins; lia. }

  generalize ((tid_sb WF_G) (fwl t (d' :: L') 0 x)  (fwl t (d :: L) 0 x) A).
  unfolder; ins; desf.
  - basic_solver.
  - right; apply sb_in_hb; done.
  - exfalso; eapply MAX; basic_solver 12.
  - left.
  remember (fwl t (d :: L) 0 x) as e.
  remember (fwl t (d' :: L') 0 x) as e'.
  destruct e, e'; desf.
  assert (x0 = loc (fwl t (d :: L) 0 x)).
  { rewrite <- Heqe. vauto. } 
  assert (x1 = loc (fwl t (d' :: L') 0 x)).
  { rewrite <- Heqe'. vauto. }
  subst. 
  f_equal.
  red in FWL; desc.
  erewrite WL_LOC; [|ins;lia].
  red in FWL'; desc.
  by erewrite WL_LOC0; [|ins;lia]. }

desc.
specialize (FWL _ _ IN_MIN); desc.
red in FWL; desc.
assert (Z : 0 < length ol_min).
{ destruct ol_min; ins.
exfalso; eapply WF_S; eauto. lia. }

specialize (WL_TID 0 Z x); ins.
specialize (WL_LOC 0 Z x); ins.
specialize (WL_VAL 0 Z x); ins.
specialize (WL_RMW 0 Z x); ins.
assert (ZZ : event_in_wl (fwl t ol_min) (length ol_min) w).
{ red. eexists _,_; split; eauto.
  red. eauto. }
specialize (WL_W w ZZ).
specialize (WL_G w ZZ).
destruct ol_min as [|d ol_min]; [by ins;lia|].
clear Z.
assert (D : d x = RT tw v u) by eauto.
simpl in *.
rewrite D in *.
forward (eapply exists_add' with (w:=w) (l:= Aload x v) (t:=t)); subst; try edone.
intro ADD; desc; exists G'; splits.
- remember (fwl t (d :: ol_min) 0 x) as w.
  econs; eexists _,_,w; splits; eauto; ins.
  intro. red in FWL0; desc.
  eapply (WL_D w); split; vauto.
- red in MATCH; desc.
  red; splits; try done.
  by eapply add_preserves_wf; try edone.
  subst; ins; destruct (classic (p = t)) as [X|NEQ]; try subst p.
  * exists (fwl t ol); eapply GsimL_read_helper_t; eauto.
    destruct ol; ins; try lia.
    eby exfalso; eapply NON_EMPTY_OL.
    by eapply FWL_COPY.
  * subst. rewrite <- WL_LOC in ADD at 1.
    exists (fwl p ol); eapply GsimL_read_helper_p; eauto.
    by apply FWL_COPY.
Qed.
