(******************************************************************************)
(** * SRAG simulates SRAL: write step   *)
(******************************************************************************)

Require Import List.
Import ListNotations.

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Require Import AuxRel.
Require Import Basic.
Require Import Labels Events Execution.
Require Import SRALn Wmax Add_max SRAG.
Require Import Write_lists Match_SRAn.

Require Import FG.

Set Implicit Arguments.

Definition max_rel (rel : relation Event) e1 e2 :=
  ifP (rel e1 e2) then e2 else e1.

Definition make_wl (wl_p wl_t : nat -> Loc -> Event) (i : nat) 
  ne xe (rel : relation Event) :=  fun n => ifP (n < i) then (wl_p n)
   else (fun y => ifP (y = xe) then ne else 
             max_rel rel (wl_p n y) (wl_t (n - i) y)).





Lemma GsimL_write_helper (st st' : State) (WF_S: wf_state st')
          (G : execution) 
          t x v (SRAL: SRAL_write_step_helper st st' t x v)
(WF_ST: wf_state st)
  ( WF_G : Wf G ) 
fwl
(FWL: forall p ol (IN: In ol (st p)), 
    SGmatch_wl G p ol (fwl p ol))


          n w G' l
          (MAX: w_max G x w) (ADD: add G G' t n l w)
          (Wl: is_w_l l) (Wloc: loc_l l = x) (Wval: valw_l l = v) 

        (Rl: is_r_l l -> forall ol (IN: In ol (st t)), fwl t ol 0 x = w  )         :
    SGmatch st' G'.
Proof.
apply wf_st_to_SGmatch.
- 

eapply SRAL_write_step_helper_preserves_wf with (st:=st); try edone.
-  eapply add_preserves_wf; try edone.

- intros p ol' IN.
red in SRAL; desc. 
unfold SRAL_write_step_helper_p in *; desc.

  specialize (P _ _ IN); desf.
unfold update_ol in *; desf.

assert (LIA_HELPER : forall n1 n2, n1 <= n2 -> n2 < length ol -> n1 < length ol) by lia.

assert (FWLt := FWL _ _ P2).
specialize (FWL _ _ P1); desc.
remember (fwl t (skipn i ol)) as wl_t.
remember (fwl p ol) as wl.
remember (ThreadEvent t n l) as ne.

assert (LOC_HELPER : forall n y, n < length ol /\ 
 y <> loc_l l -> ~ is_loc (loc_l l) (wl n y)).
{ ins; desc. intro X.
unfold is_loc in X.
apply H0.
rewrite <- X. symmetry. eapply FWL. auto. } 

assert (LOC_HELPER_t : forall n y, n < length ol - i /\ 
 y <> loc_l l -> ~ is_loc (loc_l l) (wl_t n y)).
{ ins; desc. intro X.
unfold is_loc in X.
apply H0.
rewrite <- X. symmetry. eapply FWLt.
rewrite skipn_length; lia. } 

destruct (classic (p = t)); try subst p.
+ assert (II: i = 0) by auto. subst i.
ins; clear P0.
clear FWLt.
red in FWL; desc.
remember (fun n => fun y => ifP (y = loc_l l) then ne else (wl n y)) as new_wl.

assert (EV: event_in_wl new_wl (length ol) ⊆₁ 
  (event_in_wl wl (length ol) ∩₁ (fun e => ~ is_loc (loc_l l) e)) ∪₁ (eq ne)).
{ unfold event_in_wl, in_wl. unfolder; ins; desf; eauto.
  left. splits; eauto 10. }
 
exists new_wl.
red; splits; autorewrite with calc_length in *; ins.
{ red; splits; autorewrite with calc_length in *; ins.
- rewrite EV.
unionL.
red in WL_WF; desc.
rewrite WL_W; basic_solver.
basic_solver.
-  
rewrite EV.
rewrite add_acts_set; try edone. 
unionL.

unionR left.
red in WL_WF; desc.
rewrite WL_G; basic_solver.
basic_solver.
- erewrite nth_map2 with (b := init_ps); try lia.
unfold upd_ps, upd.
desf; ins.
apply WL_WF.
lia.
- 
desf; try lia.
by apply WL_WF.
- erewrite nth_map2 with (b := init_ps); try lia.
unfold upd_ps, upd.
desf; ins.
apply WL_WF.
lia.

- erewrite nth_map2 with (b := init_ps); try lia.

unfold upd_ps, upd.
desf; ins; split; ins; eauto.
by eapply w_max_add_write; try edone. 
+ eapply w_max_add_neq_loc; try edone.
eby eapply wf_wl_last_max.
intro X.
eapply LOC_HELPER; try edone. 
unfold is_loc. rewrite <- X. apply MAX.
+ apply WL_WF; try lia.
by rewrite (add_co_monotone ADD). }

red; splits.
*
rewrite EV; clear EV.
rewrite add_max_write_dom_co_hb; try edone.
relsf; splits.
-- 
red in WL_CONS; desc.
revert WL_D; basic_solver 12.
--
rewrite (wf_coE WF_G).
by unfolder; ins; desf; eapply add_fresh; try edone.
-- remember (hb G)^? as chb.
unfold event_in_wl, in_wl.
unfolder; ins; desf.
rename y into w.
assert (AA: fwl t ol 0 (loc_l l) = w).
by eapply Rl; auto.
red in WL_CONS; desc.
eapply WL_O.
exists (fwl t ol n0 x0); splits; eauto; [|basic_solver].
apply event_order_in_wl_k; try lia.

-- rewrite (wf_coE WF_G).
 unfolder; ins; desc; subst.
eapply add_fresh; try edone.
-- basic_solver 21.
-- unfolder; ins; desc; subst.
eapply add_fresh; try edone.
* 
rewrite add_max_write_co_hb; try edone.

assert (EVO: event_order_in_wl new_wl (length ol) ⊆ 
  event_order_in_wl wl (length ol) ∩ ((fun e => ~ is_loc (loc_l l) e) × (fun e => ~ is_loc (loc_l l) e)) ∪
  (event_in_wl wl (length ol) ∩₁ (fun e => ~ is_loc (loc_l l) e)) × (eq ne) ∪
  (eq ne) × (event_in_wl wl (length ol) ∩₁ (fun e => ~ is_loc (loc_l l) e)) ∪
  (eq ne) × (eq ne)).
{ unfold event_order_in_wl, worder_wl, event_in_wl, in_wl.
unfolder; ins; desf; eauto 15.
by left; left; left; splits; eauto 15. }

rewrite EVO; clear EVO.


relsf. unionL.
-- red in WL_CONS; desc.
revert WL_O. basic_solver 12.

-- rewrite (wf_coE WF_G).
generalize (add_fresh ADD).
basic_solver 21.

-- rewrite (wf_coE WF_G), (wf_hbE WF_G).
generalize (add_fresh ADD).
basic_solver 21.

-- rewrite (wf_coE WF_G).
generalize (add_fresh ADD).
basic_solver 21.

-- unfold is_loc. unfolder; ins; desf. 


-- rewrite (wf_coE WF_G).
generalize (add_fresh ADD).
basic_solver 21.

-- remember (hb G)^? as chb.
unfolder; ins; desf.
{ red in WL_CONS; desc.
eapply WL_D.
revert H4; 
basic_solver 12. }

revert H2.
unfold event_in_wl, in_wl.
unfolder; ins; desf.
rename y into w.
assert (AA: fwl t ol 0 (loc_l l) = w).
by eapply Rl; auto.
red in WL_CONS; desc.
eapply WL_O.
exists (fwl t ol n0 x); splits; eauto; [|basic_solver].
apply event_order_in_wl_k; try lia.


-- rewrite (wf_coE WF_G).
generalize (add_fresh ADD).
basic_solver 21.

+ (* p <> t *) 
clear P0.
red in FWL; desc.
red in FWLt; desc.
remember (make_wl wl wl_t i ne (loc_l l) G.(co)) as new_wl.

cdes WL_WF0.
rename WL_W into WL_W_t.
rename WL_G into WL_G_t.
rename WL_TID into WL_TID_t.
rename WL_LOC into WL_LOC_t.
rename WL_VAL into WL_VAL_t.
rename WL_RMW into WL_RMW_t.
rename WL_CONS0 into WL_CONS_t.

rewrite List.skipn_length in *.

assert (EV: event_in_wl new_wl (length ol) ⊆₁ 
event_in_wl wl (length ol) ∪₁
(event_in_wl wl_t (length ((skipn i ol))) ∩₁ (fun e => ~ is_loc (loc_l l) e)) ∪₁ (eq ne)).
{subst. unfold event_in_wl, in_wl, make_wl, max_rel. unfolder; ins; desf; eauto 12.

left. right. splits.
(* assert (length ol > 1). *)

exists (fwl t (skipn i ol) (n0 - i)),x0; splits; eauto.
exists (n0-i); splits.
rewrite List.skipn_length.
 lia. done.
unfold is_loc; intros X.
rewrite <- X in n2.
apply n2.
symmetry.
eapply WL_LOC_t.
lia.
}

exists new_wl.
red; splits; red; splits; ins; unfold update_ol;
 autorewrite with calc_length in *; ins.
all: replace (Init.Nat.min i (length ol)) with i by lia.
all: rewrite List.skipn_length in *.
all: replace (i + (length ol - i)) with (length ol) by lia.

*  rewrite EV.

unionL.
red in WL_WF; desc.
rewrite WL_W; basic_solver.
rewrite WL_W_t; basic_solver.
basic_solver.


* rewrite EV.
rewrite add_acts_set; try edone. 

unionL.
red in WL_WF; desc.
rewrite WL_G; basic_solver.
rewrite WL_G_t; basic_solver.
basic_solver.

* subst. unfold make_wl, max_rel, upd_read_ps, upd_ps, upd.

rewrite !nth_app; autorewrite with calc_length in *; ins.
desf; try lia; ins.
all: rewrite !nth_map2 with (b:=init_ps).
all: try by rewrite firstn_length; lia.
all: try by rewrite skipn_length; lia.
all: desf; ins.
all: rewrite ?nth_firstn; try lia.
all: try by apply WL_WF; lia.
all: replace (Init.Nat.min i (length ol)) with i by lia.
apply WL_TID_t; lia.
rewrite skipn_nth.
replace (i + (n0 - i)) with n0 by lia.
apply WL_WF; lia.


* subst. unfold make_wl, max_rel, upd_read_ps, upd_ps, upd.
desf; try lia; ins.
apply WL_WF; lia.
apply WL_LOC_t; lia.
apply WL_WF; lia.


* subst. unfold make_wl, max_rel, upd_read_ps, upd_ps, upd.

rewrite !nth_app; autorewrite with calc_length in *; ins.
desf; try lia; ins.
all: rewrite !nth_map2 with (b:=init_ps).
all: try by rewrite firstn_length; lia.
all: try by rewrite skipn_length; lia.
all: desf; ins.
all: rewrite ?nth_firstn; try lia.
all: try by apply WL_WF; lia.
all: replace (Init.Nat.min i (length ol)) with i by lia.
apply WL_VAL_t; lia.
rewrite skipn_nth.
replace (i + (n0 - i)) with n0 by lia.
apply WL_WF; lia.

* subst. unfold make_wl, max_rel, upd_read_ps, upd_ps, upd.

rewrite !nth_app; autorewrite with calc_length in *; ins.
desf; try lia; ins.
all: rewrite !nth_map2 with (b:=init_ps).
all: try by rewrite firstn_length; lia.
all: try by rewrite skipn_length; lia.
all: desf; ins.
all: rewrite ?nth_firstn; try lia.
all: split; ins; desf.
all: replace (Init.Nat.min i (length ol)) with i in * by lia.
all: try by eapply w_max_add_write; try edone.
all: try apply WL_WF; try lia.
all: try apply WL_RMW_t; try lia.
all: try by rewrite (add_co_monotone ADD).

-- eapply w_max_add_neq_loc; try edone.
eapply wf_wl_last_max; try edone.
rewrite skipn_length. lia. lia.
rewrite WL_LOC_t; try lia.
red in MAX; unfolder in MAX; desc.
congruence.

-- eapply w_max_add_neq_loc; try edone.
eapply wf_wl_last_max; try edone.
lia.
rewrite skipn_nth in H0.
rewrite <- H0; f_equal; f_equal; try lia.
red in WL_WF; desc. 
rewrite WL_LOC; try lia.
red in MAX; unfolder in MAX; desc.
congruence.

-- apply WL_RMW_t; try lia.
rewrite skipn_nth.
replace (i + (n0 - i)) with n0 by lia.
apply WL_WF. lia.
rewrite (add_co_monotone ADD).
done.

-- exfalso.
eapply H0 with (b := (ThreadEvent t n l)).
eapply add_max_write_co; try edone.
right; unfolder; splits; eauto.
apply WL_WF. apply event_in_wl_k. lia.
apply WL_WF. apply event_in_wl_k. lia.
apply WL_WF. lia.

-- eapply w_max_add_neq_loc; try edone.
eapply wf_wl_last_max; try edone.
lia.
red in WL_WF; desc. 
rewrite WL_LOC; try lia.
red in MAX; unfolder in MAX; desc.
congruence.


* 
red in WL_CONS; desc.

rewrite add_max_write_dom_co_hb; try edone.
subst.
unfold event_in_wl, in_wl, make_wl, max_rel. 
remember (hb G)^? as chb.
unfolder; ins; desf; try lia.
all: try by eapply WL_D; unfolder; splits; eauto 12;
apply event_in_wl_k; lia.
-- by apply (wf_coE WF_G) in H1;
unfolder in H1; ins; desc; subst;
eapply add_fresh; edone.
-- eapply WL_D with (x:= (fwl p ol n0 x0)).
generalize (co_trans WF_G); unfolder; splits; eauto 12.
apply event_in_wl_k; lia.

* rewrite add_max_write_co_hb; try edone.
subst.
unfold event_order_in_wl, worder_wl, make_wl, max_rel.
rewrite <- !seqA.
remember ((co G);; (hb G)^?) as chb.
unfolder; ins; desc.

assert (DCO: ~ dom_rel (co G)  (ThreadEvent t n l)).
{  unfolder; ins; desc.
intro X. desc.
apply (wf_coE WF_G) in X.
unfolder in X; desc.

eapply add_fresh; eauto 12. }

assert (CCO: ~ codom_rel (co G)  (ThreadEvent t n l)).
{  unfolder; ins; desc.
intro X. desc.
apply (wf_coE WF_G) in X.
unfolder in X; desc.

eapply add_fresh; eauto 12. }

assert (CHB: ~ codom_rel (hb G)  (ThreadEvent t n l)).
{  unfolder; ins; desc.
intro X. desc.
apply (wf_hbE WF_G) in X.
unfolder in X; desc.

eapply add_fresh; eauto 12. }

assert (forall a b c d, a <= b -> b <= c -> c < d -> a < d).
lia.

assert (forall j, j < length ol -> 
  is_loc (loc_l l) (fwl p ol j y) -> 
y <> loc_l l -> 
False).
{ intros j A B C.
apply C.
unfold is_loc in B.
rewrite <- B.
symmetry.
by apply WL_WF. }

assert (n1 < length ol) by lia.
assert (n1 <= n2 ) by lia.
assert (n1 - i < length ol - i) by lia.
assert ((n1 - i) <= (n2 - i) ) by lia.
assert ((n2 - i) < length ol - i) by lia.
(*   *)

assert 
 (H1': exists z0 : Event, chb z x \/
     (((exists y z0 : Event, chb z y /\ is_tid t y) \/
       (acts_set G z /\ W z) /\ is_loc (loc_l l) z) \/
      (exists y z0 : Event, chb z y /\ w = y) /\
      is_r_l l) /\ ThreadEvent t n l = x).
{ subst chb. revert H1. 
generalize (co_trans WF_G).
clear.
unfolder; ins; desf; eauto 12. }
clear H1.

assert (cohb : forall a b c, co G a b -> (co G ⨾ (hb G)^?) b c ->
(co G ⨾ (hb G)^?) a c).
 { generalize (co_trans WF_G); clear; basic_solver 12. }


desf.
all: try lia.
all: try by revert H1' DCO CCO CHB; clear; basic_solver.

all: try by eapply add_fresh; try edone;
eapply WL_G_t;
unfold event_in_wl, in_wl;
basic_solver 8.

all: try by eapply add_fresh; try edone;
eapply WL_WF;
unfold event_in_wl, in_wl; 
basic_solver 8.

all: try by red in WL_CONS; desc;
eapply WL_O;
eexists; splits; [|revert H1'; basic_solver 18];
apply event_order_in_wl_k; lia.

all: try by red in WL_CONS; desc;
eapply WL_O;
eexists; splits; [|eapply cohb; eauto]; 
apply event_order_in_wl_k; lia.

all: try by red in WL_CONS_t; desc;
eapply WL_O;
eexists; splits; [|revert H1'; basic_solver 18];
apply event_order_in_wl_k; lia.


-- assert (CO: (co G)^? (fwl t (skipn i ol) (n2 - i) y) (fwl p ol n2 y)).
      { eapply tot_ex'.
        eapply WF_G.
        + unfolder; splits;
apply WL_WF; unfold event_in_wl, in_wl; basic_solver 12.
        + 
unfolder; splits.
apply WL_G_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_W_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_LOC_t; unfold event_in_wl, in_wl; basic_solver 12.

        + done.
      }

assert (CO' : (co G ⨾ (hb G)^?) (fwl t (skipn i ol) (n2 - i) y) (fwl t (skipn i ol) (n1 - i) x0)).
{  revert H1' CO.
generalize (co_trans WF_G).
clear.
unfolder; ins ; desf; eauto 20.
setoid_rewrite CO; 
basic_solver 20.
setoid_rewrite CO; 
basic_solver 20.
}


red in WL_CONS_t; desc.
eapply WL_O.
unfold event_order_in_wl, worder_wl.


revert CO'.
basic_solver 20.




-- 
red in WL_CONS_t; desc.
eapply WL_D.
unfold event_in_wl, in_wl.
revert H1'.

basic_solver 18.

--


assert (CO: (co G)^? (fwl t (skipn i ol) (n2 - i) y) (fwl p ol n2 y)).
      { eapply tot_ex'.
        eapply WF_G.
        + unfolder; splits;
apply WL_WF; unfold event_in_wl, in_wl; basic_solver 12.
        + 
unfolder; splits.
apply WL_G_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_W_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_LOC_t; unfold event_in_wl, in_wl; basic_solver 12.

        + done.
      }

assert (CO' : (co G ⨾ (hb G)^?) (fwl t (skipn i ol) (n2 - i) y) y0).
{ remember (fwl t (skipn i ol) (n2 - i) y) as a. revert H1' CO; generalize (co_trans WF_G); clear. 
basic_solver 12. }

clear H1'.



red in WL_CONS_t; desc.
eapply WL_D.

split.
eapply event_in_wl_k with (k:=n2 - i); try lia.
revert CO'; basic_solver 18.

--
apply n4.
unfold is_loc in *.
rewrite WL_LOC_t in *; try lia.
congruence.
-- 
apply n4.
unfold is_loc in *.
red in WL_WF; desc.
rewrite WL_LOC in *; try lia.
congruence.


-- 
clear c.



rename y0 into w.
assert (AA: fwl t (skipn i ol) 0 (loc_l l) = w).
by apply Rl; auto.
red in WL_CONS_t; desc.
eapply WL_O with (x:=w).
exists (fwl t (skipn i ol) (n2 - i) y); splits; eauto.
by desf; apply event_order_in_wl_k; try lia.

-- 


assert (CO: (co G)^? (fwl t (skipn i ol) (n2 - i) y) (fwl p ol n2 y)).
      { eapply tot_ex'.
        eapply WF_G.
        + unfolder; splits;
apply WL_WF; unfold event_in_wl, in_wl; basic_solver 12.
        + 
unfolder; splits.
apply WL_G_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_W_t; unfold event_in_wl, in_wl; basic_solver 12.
apply WL_LOC_t; unfold event_in_wl, in_wl; basic_solver 12.

        + done.
      }

assert (CO' : (co G ⨾ (hb G)^?) (fwl t (skipn i ol) (n2 - i) y) y0).
{ remember (fwl t (skipn i ol) (n2 - i) y) as a. revert H1' CO; generalize (co_trans WF_G); clear. 
basic_solver 12. }

clear H1'.

rename y0 into w.
assert (AA: fwl t (skipn i ol) 0 (loc_l l) = w).
by apply Rl; auto.
red in WL_CONS_t; desc.
eapply WL_O with (x:=w).
exists (fwl t (skipn i ol) (n2 - i) y); splits; eauto.
by desf; apply event_order_in_wl_k; try lia.

Qed.

Lemma GsimL_write (st st' : State) (G : execution)
      t x v (SRAL: SRAL_step st st' (Some (t, (Astore x v))))
      (MATCH: SGmatch st G) : 
         exists G', SRAG_step G G' (Some (t, (Astore x v))) /\ 
                    SGmatch st' G'.
Proof.
inversion SRAL; subst; clear SRAL.
assert (WF_G: Wf G) by apply MATCH.
forward (eapply (max_w_exists WF_G x)); intro MAX; desc.
forward (eapply exists_add' with (w:=w) (l:= Astore x v) (t:=t)); try edone.
all: try by red in MAX; revert MAX; basic_solver.
ins; desf.
exists G'; splits.
red; eauto 21.
cdes MATCH.
apply SGmatch_SGmatch_fun in MATCH; desc.
eapply GsimL_write_helper; try edone.
eapply SRAL_write_step_helper_preserves_wf; try edone.
eapply SGmatch_implies_wf_st; try edone.
eapply SGmatch_implies_wf_st; try edone.
Qed.
