(******************************************************************************)
(** * SRAL <-> SRAG simulation definitions   *)
(******************************************************************************)

Require Import List.
Import ListNotations.

Require Import Coq.Logic.ClassicalChoice.

Require Import PeanoNat Lt Gt.
Require Import Lia.
Require Import Hahn.

Require Import Basic.
Require Import Labels Events Execution.
Require Import SRALn SRAG.

Require Import FG.
Require Import Wmax.
Require Import Write_lists.

Set Implicit Arguments.

Section Match.

Definition wf_wl (wl : nat -> Loc -> Event) (ol : list PStore) (G : execution) :=
  ⟪ WL_W  : event_in_wl wl (length ol) ⊆₁ is_w ⟫ /\
  ⟪ WL_G  : event_in_wl wl (length ol) ⊆₁ acts_set G ⟫ /\
  ⟪ WL_TID  : forall n (LEN: n < length ol) x, 
              tid (wl n x) = tid_opt (nth n ol init_ps x) ⟫ /\
  ⟪ WL_LOC  : forall n (LEN: n < length ol) x, 
              loc (wl n x) = x ⟫ /\
  ⟪ WL_VAL  : forall n (LEN: n < length ol) x, 
              valw (wl n x) = val_opt (nth n ol init_ps x) ⟫ /\
  ⟪ WL_RMW  : forall n (LEN: n < length ol) x,
               (rmw_flag_opt (nth n ol init_ps x) = true) <->
                max_elt G.(co) (wl n x) ⟫.
(*   ⟪ WL_LAST  : forall x, max_elt G.(co) (wl (pred (length ol)) x) ⟫. *)

Lemma wf_wl_last_max wl ol G (WF_WL : wf_wl wl ol G) x n
(LEN: n < length ol) (RMW: rmw_flag_opt (nth n ol init_ps x) = true) :
w_max G x (wl n x).
Proof.
red; unfolder; splits.
all: apply WF_WL; unfold event_in_wl, in_wl; eauto 12.
Qed.

Definition consistent_wl (wl : nat -> Loc -> Event) N G p :=
  ⟪ WL_D  : (event_in_wl wl N) ∩₁ dom_rel ( G.(co) ⨾ (hb G)^? ⨾ ⦗ is_tid p ⦘) ⊆₁ ∅ ⟫ /\
  ⟪ WL_O  : irreflexive ((event_order_in_wl wl N) ⨾ G.(co) ⨾ (hb G)^?) ⟫.

Definition SGmatch_wl G p ol wl :=
  ⟪ WL_WF : wf_wl wl ol G ⟫ /\ 
  ⟪ WL_CONS : consistent_wl wl (length ol) G p ⟫.

Definition SGmatch (st : State) (G : execution) :=
(*   ⟪ WF_ST : wf_state st ⟫ /\ *)
  ⟪ NON_EMPTY_POT : forall t, st t <> [] ⟫ /\
  ⟪ NON_EMPTY_OL  : forall t L (IN: In L (st t)), L <> [] ⟫ /\
  ⟪ RMW_END : forall t L (IN: In L (st t)) y,
    rmw_flag_opt (nth (Init.Nat.pred (length L)) L init_ps y) = true ⟫ /\
  ⟪ WF_G : Wf G ⟫ /\
  ⟪ SMATCH: forall p ol (IN: In ol (st p)), 
           exists wl, SGmatch_wl G p ol wl ⟫.

Lemma wf_st_to_SGmatch st G (WF_ST : wf_state st) 
  (WF_G : Wf G)
  (SMATCH: forall p ol (IN: In ol (st p)), 
           exists wl, SGmatch_wl G p ol wl) :
 SGmatch st G.
Proof.
red; splits; eauto; try apply WF_ST.
ins.
red in WF_ST; desc.
erewrite SAME_END; try edone.
Qed.

Lemma SGmatch_implies_wf_st st G (MATCH: SGmatch st G) :
  wf_state st.
Proof.
red in MATCH; desc.
red; splits; eauto.

- ins.
specialize (SMATCH _ _ IN); desc.
red in SMATCH; desc.
red in WL_WF; desc.
assert (N2: n2 < length L) by eby eapply nth_error_Some'.
assert (N1: n1 < length L) by lia.

eapply nth_error_nth in IN2.
rewrite <- IN2.
eapply WL_RMW; try done.
unfolder; ins.

eapply nth_error_nth in IN1.
rewrite <- IN1 in RMW1.
eapply WL_RMW in RMW1; try done.
eapply RMW1 with (b:=b).

assert (~ (co G)^? b (wl n1 y)).
{ intro X.
red in WL_CONS; desc. 
eapply WL_O with (x:= (wl n1 y)).
unfold event_order_in_wl, worder_wl.
unfolder; ins; desf.
exists (wl n2 y); splits.
eexists _,_,_,_; splits; try edone.
exists n1, n2; splits; eauto 10; try lia.
exists (wl n1 y); splits; eauto.
revert X; generalize (@co_trans G); basic_solver. }


eapply tot_ex.
+ apply WF_G.
+ unfolder; splits.
apply (wf_coE WF_G) in REL; revert REL; basic_solver.
apply (wf_coD WF_G) in REL; revert REL; basic_solver.
apply (wf_col WF_G) in REL.
unfold same_loc in *. rewrite <- REL.
apply WL_LOC; done.
+ unfolder; splits.
apply WL_G; apply event_in_wl_k; try done.
apply WL_W; apply event_in_wl_k; try done.
apply WL_LOC; done.
+ revert H; basic_solver.
+ revert H; basic_solver.
- 

assert (exists f, forall x, w_max G x (f x)).
{ apply choice.
apply max_w_exists; done. }

desc.
exists (fun x => valw (f x)), (fun x => tid (f x)).
ins.

specialize (RMW_END _ _ IN y).
specialize (SMATCH _ _ IN); desc.
red in SMATCH; desc; cdes WL_WF.
assert (LENH: (Init.Nat.pred (length L)) < length L).
{ destruct L; ins; try lia. }
specialize (WL_TID _ LENH y).
specialize (WL_VAL _ LENH y).
specialize (WL_LOC _ LENH y).
specialize (WL_RMW _ LENH y).
assert (HC := RMW_END).
apply WL_RMW in RMW_END.
replace (f y) with (wl (Init.Nat.pred (length L)) y); cycle 1.
{ eapply max_w_unique; try edone; auto.
red; unfolder; splits; eauto.
 apply WL_WF; unfold event_in_wl, in_wl; eauto 12.
 apply WL_WF; unfold event_in_wl, in_wl; eauto 12. }
rewrite WL_TID.
rewrite WL_VAL.

destruct (nth (Init.Nat.pred (length L)) L init_ps y) eqn:E.
f_equal; ins.
Qed.


Lemma SGmatch_SGmatch_fun st G (A: SGmatch st G): 
 exists fwl,  forall p ol (IN: In ol (st p)), 
    wf_wl (fwl p ol) ol G /\ 
    consistent_wl (fwl p ol) (length ol) G p.
Proof.
  intros; red in A; desc.
  apply (choice (fun p fwlp => forall (ol : list PStore),
    In ol (st p) ->
    wf_wl (fwlp ol) ol G /\ consistent_wl (fwlp ol) (length ol) G p)).
  intro p.
  apply (choice (fun ol yol => 
    In ol (st p) ->
    wf_wl yol ol G /\ consistent_wl yol (length ol) G p)); intro ol.
  destruct (classic (In ol (st p))).
  - edestruct SMATCH; eauto.
  - exists (fun _ => fun _ => default_event); contradiction.
Qed.

(******************************************************************************)
(** * total relation   *)
(******************************************************************************)

Lemma SGmatch_total G (WF_G: Wf G) :
  exists BB, SGmatch BB G.
Proof.
assert (exists f : Loc -> Event, forall x, w_max G x (f x)).
{  eapply choice. by apply max_w_exists. }
desc.
exists (fun t => [[fun x => RT (tid (f x)) (valw (f x)) true]]).
red; splits; auto.
- ins; desf.
- ins; try red; splits; desf; ins; desf.
- ins; desf; ins; desf.
- ins. exists (fun _ => fun x => f x).
unfold w_max in H; unfolder in H.
red; ins; splits; red; splits; ins; desf; simpl in *; desf; ins; desf; try lia.
+ red. unfold event_in_wl, in_wl. ins. desf. specialize (H x0); desf.
+ red. unfold event_in_wl, in_wl. ins. desf. specialize (H x0); desf.
+ specialize (H x); desf.
+ specialize (H x); desf.
+ unfold event_in_wl, in_wl.
  unfolder; ins; desf; specialize (H x0); desf; eauto.
+ unfold event_order_in_wl, worder_wl, in_wl.
  unfolder; ins; desf; try lia; specialize (H y); desf; eauto.
Qed.

(******************************************************************************)
(** * lemmas for initialization   *)
(******************************************************************************)

(* Lemma SGmatch_init  (st : State) (INIT : SRALinit st) : 
  SGmatch st Ginit.
Proof.
unfold SRALinit in *; desf.
red; ins; splits; ins; desf.
- by apply Ginit_wf.
- exists (fun n => fun x => InitEvent x).
red; ins; splits; red; splits; ins; ins; desf; try basic_solver; try lia.
+ unfolder; unfold event_in_wl, in_wl; ins; desc; subst x; type_solver.
+ unfolder; unfold event_in_wl, in_wl; ins; desc; subst.
  apply Ginit_acts_set; ins.
Qed. *)

Lemma SGmatch_init  (st : State) (INIT : SRALinit st) : 
  SGmatch st Ginit.
Proof.
unfold SRALinit in *; desf.
red; ins; splits; ins; desf; try specialize (INIT t); desc; auto.
- specialize (INIT _ IN).
eapply Forall_nth in INIT.
rewrite <- INIT; desf.
destruct L; try (ins; lia).
exfalso; eapply NON_EMPTY_OL; eauto.
- by apply Ginit_wf.
- exists (fun n => fun x => InitEvent x).
red; ins; splits; red; splits; ins; ins; desf; try basic_solver; try lia.
+ unfolder; unfold event_in_wl, in_wl; ins; desc; subst x; type_solver.
+ unfolder; unfold event_in_wl, in_wl; ins; desc; subst.
  apply Ginit_acts_set; ins.
+ specialize (INIT p); desc.
specialize (INIT _ IN).
eapply Forall_nth in INIT.
rewrite <- INIT; desf.
auto.
+ specialize (INIT p); desc.
specialize (INIT _ IN).
eapply Forall_nth in INIT.
rewrite <- INIT; desf.
auto.
+ unfold max_elt.
split; eauto.
intros _.
specialize (INIT p); desc.
specialize (INIT _ IN).
eapply Forall_nth in INIT.
rewrite <- INIT; desf.
auto.
Qed.

Lemma SGmatch_G_init st (WF_S: wf_state st) (MATCH: SGmatch st Ginit) : 
  SRALinit st.
Proof.
red; ins; splits; ins.
apply MATCH.
eapply MATCH; try edone.
apply ForallE.
ins.
forward eapply In_nth with (d:=init_ps); try edone; ins; desf.
symmetry.

rename H0 into LEN.
red in MATCH; desc.
specialize (SMATCH _ _ IN); desc.
red in SMATCH; desc.
red in WL_WF; desc.
apply functional_extensionality; ins.
specialize (WL_TID _ LEN x).
specialize (WL_VAL _ LEN x).
specialize (WL_RMW _ LEN x).
destruct (nth n L init_ps x) eqn:E; ins.

rewrite <- WL_TID.
rewrite <- WL_VAL.
assert (K : rmw_flag = true).
{ apply WL_RMW. unfold max_elt; ins; desf. }
rewrite K.

specialize (WL_G (wl n x)).
forward (apply WL_G).
by apply event_in_wl_k.

intro A.
apply Ginit_acts_set in A.
destruct (wl n x); vauto.
Qed.

End Match.

