(******************************************************************************)
(** * Main result: SRAL and SRAG have the same runs     *)
(******************************************************************************)

Require Import List.
Import ListNotations.

Require Import Hahn.

Require Import Labels Events Execution FG.
Require Import SRALn SRAG.
Require Import Match_SRAn.

Require Import GsimL_read_SRAn GsimL_write_SRAn GsimL_rmw_SRAn GsimL_silent_SRAn.
Require Import LsimG_read_SRAn LsimG_write_SRAn LsimG_rmw_SRAn.

Require Import Memory_system.

Set Implicit Arguments.

Lemma SRAG_reachable_wf G
  (REACH: reachable SRAGinit SRAG_step G): Wf G.
Proof.
eapply MS_reachable_preserves; try edone.
- ins; unfold SRAGinit in *; subst; apply Ginit_wf.
- eby ins; eapply SRAG_step_preserves_wf.
Qed.

Lemma SRAG_reachable_consistent G
  (REACH: reachable SRAGinit SRAG_step G): sra_consistent G.
Proof.
cut (Wf G /\ sra_consistent G); try tauto.
pattern G.
eapply MS_reachable_preserves; try edone.
- ins; unfold SRAGinit in *; subst.
splits. 
by apply Ginit_wf.
by apply Ginit_sra_consistent.
- ins; desc; splits.
 eby ins; eapply SRAG_step_preserves_wf.
destruct o; [| by inversion H0; desf].
destruct p.
eapply SRAG_step_preserves_consistency; try edone.
pattern G.
eapply MS_reachable_preserves; try edone.
ins; unfold SRAGinit in *; subst.
apply H.
Qed.


Definition SRAL_run := MS_run SRALinit SRAL_step.
Definition SRAG_run := MS_run SRAGinit SRAG_step.

Lemma SRAL_in_SRAG : forall L, SRAL_run L -> SRAG_run L.
Proof.
eapply MS1_runs_in_MS2_runs_fwd with (sim_rel := SGmatch).
- ins; exists Ginit; splits.
  by vauto.
  by apply SGmatch_init.
- ins; destruct o.
  rename st2 into G.
  + left; destruct p, l.
    * eapply GsimL_read; try edone.
      red in H0; desc; auto.
    * eapply GsimL_write; try edone.
    * eapply GsimL_rmw; try edone.
  + right; splits; auto.
    eby eapply GsimL_silent.
Qed.

Lemma SRAG_in_SRAL : forall L, SRAG_run L -> SRAL_run L.
Proof.
eapply MS1_runs_in_MS2_runs_bwd with 
  (sim_rel := fun x y => SGmatch y x).
- intro G; ins. 
  apply SGmatch_total.
  eapply MS_reachable_preserves; try edone.
  + ins; unfold SRAGinit in *; subst; apply Ginit_wf.
  + eby ins; eapply SRAG_step_preserves_wf.
- ins; unfold SRAGinit in *; subst.
 exists st2; splits; vauto.
apply SGmatch_G_init; try done.
eby eapply SGmatch_implies_wf_st.
- ins; destruct o.
+ rename st1 into G. 
rename st1' into G'.
rename st2' into st'.
left; destruct p.
eexists _,_; splits; eauto.
destruct l.
    * 
 eapply LsimG_read in H; try edone; desc.
   --   exists st; splits; try done.
replace ((t, Aload x v) :: nil) 
  with ([(t, Aload x v)] ++ []) by ins.
eapply trans. econs; eauto. econs; eauto.
-- eby eapply SGmatch_implies_wf_st.
-- eby eapply SRAG_reachable_wf.
-- eby eapply SRAG_reachable_consistent.
    * forward eapply LsimG_write; try edone; ins; desf.
exists st. vauto.
    * forward eapply LsimG_rmw; try edone; ins; desf.
exists st. vauto.
  + inversion H; desf.
Qed.

Lemma SRAL_eq_SRAG : forall L, SRAL_run L <-> SRAG_run L.
Proof.
ins; split; auto using SRAL_in_SRAG, SRAG_in_SRAL.
Qed.
