(******************************************************************************)
(** * definition of w_max and some useful properties of it   *)
(******************************************************************************)

Require Import Lia.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution.

Set Implicit Arguments.

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Section Wmax.

Definition w_max G x := 
  acts_set G ∩₁ W ∩₁ (fun w => loc w = x) ∩₁ max_elt G.(co).

(******************************************************************************)
(** * properties of w_max  *)
(******************************************************************************)

Lemma max_w_maximum G (WF: Wf G) x w wm 
  (ACTS: acts_set G w) (IS_W: W w) (LOC: loc w = x) (MAX: w_max G x wm): 
  G.(co)^? w wm.
Proof.
destruct (classic (w = wm)).
basic_solver.
eapply (wf_co_total WF) in H.
- desf; try basic_solver. red in MAX; unfolder in MAX; desf.
  exfalso; eauto.
- basic_solver.
- red in MAX; unfolder in MAX; desf.
Qed.

Lemma max_w_unique G (WF: Wf G) x w1 w2 
  (MAX1: w_max G x w1) (MAX2: w_max G x w2): 
  w1 = w2.
Proof.
destruct (classic (w1 = w2)); try done.
exfalso.
unfold w_max in *; unfolder in *; desc.
eapply (wf_co_total WF) in H; desf; eauto.
Qed.

Lemma max_w_exists G (WF: Wf G) x : 
  exists w, w_max G x w.
Proof.
assert (exists w, G.(co)＊ (InitEvent x) w /\ max_elt G.(co) w).
{ eapply last_exists with (dom := G.(acts)).
  eby eapply co_acyclic.
  intros c H.
  apply rtE in H; unfolder in H; desf.
  by apply WF.
  apply clos_trans_of_transitiveD in H; [|by apply WF].
  hahn_rewrite (dom_r (wf_coE WF)) in H; unfolder in H.
  unfold acts_set in *; desf. }
desc.
apply clos_refl_trans_of_transitive in H; [|by apply WF].

assert (acts_set G w).
{ destruct H.
by subst; eapply WF.
  hahn_rewrite (dom_r (wf_coE WF)) in H; unfolder in H.
  unfold acts_set in *; desf. }

assert (W w).
{ destruct H.
  subst; type_solver.
  by hahn_rewrite (dom_r (wf_coD WF)) in H; unfolder in H; desc. }

assert (loc w = x).
{ destruct H.
subst; type_solver.
apply WF.(wf_col) in H.
unfold same_loc in *; type_solver. }


exists w; red; ins; splits; try done.
Qed.

Lemma max_w_exists_unique G (WF: Wf G) x : 
  exists! w, w_max G x w.
Proof.
unfold unique.
forward (apply max_w_exists with (x:=x)); eauto.
ins; desf; eexists; splits; eauto.
ins; eapply max_w_unique; eauto.
Qed.

Lemma max_w_exists_unique_fun G (WF: Wf G) x :
  exists! v, forall w, w_max G x w -> v = valw w.
Proof.
ins.
generalize (max_w_exists_unique WF); intro B.
specialize (B x).
destruct B as [w B]; destruct B as [MAX U].
exists (valw w).
red; splits; try done.
* ins; splits; try done.
ins; erewrite U; edone.
* intro v'; ins; desf.
specialize (H w MAX); desf; lia.
Qed.

Lemma w_max_rewrite_helper G (WF: Wf G)  x w (MAX : w_max G x w):
⦗w_max G x⦘ ≡ ⦗eq w⦘.
Proof.
unfolder; ins; desf; splits; ins; desf; splits; eauto.
eapply max_w_unique; eauto.
Qed.

Lemma w_max_eq G x w 
  (WF: Wf G) (MAX: w_max G x w) A :
  w_max G x ⊆₁ A <-> A w.
Proof.
unfolder; split; ins; eauto.
replace x0 with w; eauto.
eapply max_w_unique; eauto.
Qed.

End Wmax.
