(******************************************************************************)
(** * definition of w_max_t and some useful properties of it   *)
(******************************************************************************)

Require Import Coq.Logic.ClassicalChoice.

Require Import Lia.

Require Import Hahn.
Require Import AuxRel.

Require Import Labels Events Execution.

Require Import FG SRAG.
Set Implicit Arguments.

Notation "'R'" := is_r.
Notation "'W'" := is_w.
Notation "'RMW'" := is_rmw.

Section Wmax_t.


Definition w_max_t G t x := 
  acts_set G ∩₁ W ∩₁ (fun w => loc w = x) ∩₁ 
  (dom_rel ((hb G)^? ⨾ ⦗is_tid t⦘) ∪₁ is_init) ∩₁ 
  (set_compl (dom_rel ( G.(co) ⨾ (hb G)^? ⨾ ⦗ is_tid t ⦘))).

(******************************************************************************)
(** * properties of w_max  *)
(******************************************************************************)


Lemma max_w_t_unique G (WF: Wf G) t x w1 w2 
  (MAX1: w_max_t G t x w1) (MAX2: w_max_t G t x w2): 
  w1 = w2.
Proof.
destruct (classic (w1 = w2)); try done.
exfalso.
unfold w_max_t in *.
unfold set_compl in *.
 unfolder in *; desc.
eapply (wf_co_total WF) in H; desf; eauto 10.

all: try by apply (wf_co_init WF) in H; unfolder in H; desf.

Qed.

Lemma max_w_t_exists G (WF: Wf G) t x : 
  exists w, w_max_t G t x w.
Proof.
assert (T: transitive (co G ⨾ ⦗dom_rel ((hb G)^? ⨾ ⦗is_tid t⦘) ∪₁ is_init⦘)).
{ generalize (co_trans WF). basic_solver 12. }


assert (exists w, (G.(co) ;; ⦗dom_rel ((hb G)^? ⨾ ⦗is_tid t⦘) ∪₁ is_init⦘) ＊ (InitEvent x) w  
/\ max_elt (G.(co) ;; ⦗dom_rel ((hb G)^? ⨾ ⦗is_tid t⦘) ∪₁ is_init⦘)  w).
{ eapply last_exists with (dom := G.(acts)).
eapply acyclic_mon.
  eby eapply co_acyclic.
basic_solver.
  intros c H.
apply rt_of_trans in H; try edone.
hahn_rewrite (dom_r (wf_coE WF)) in H.



 unfolder in H; desf.
  by apply WF.

 }
desc.
apply clos_refl_trans_of_transitive in H; try edone.

assert (acts_set G w).
{ destruct H.
by subst; eapply WF.
  hahn_rewrite (dom_r (wf_coE WF)) in H; unfolder in H.
  unfold acts_set in *; desf. }

assert (W w).
{ destruct H.
  subst; type_solver.
  hahn_rewrite (dom_r (wf_coD WF)) in H.
 unfolder in H; desf. }

assert (loc w = x).
{ destruct H.
subst; type_solver.
hahn_rewrite WF.(wf_col) in H.
unfold same_loc in *; unfolder in H. desf. }


exists w. red. unfold set_compl; unfolder. splits; try done.

unfolder in H; desf; eauto 20.
right. rewrite <- H. vauto.

intro C; desc.

eapply H0.
basic_solver 21.

Qed.

Lemma max_w_t_exists_unique G (WF: Wf G) t x : 
  exists! w, w_max_t G t x w.
Proof.
unfold unique.
forward (apply max_w_t_exists with (x:=x)); eauto.
ins; desf; eexists; splits; eauto.
ins; eapply max_w_t_unique; eauto.
Qed.

Lemma max_w_t_exists_unique_fun G (WF: Wf G) t x :
  exists! v, forall w, w_max_t G t x w -> v = valw w.
Proof.
ins.
generalize (max_w_t_exists_unique WF); intro B.
specialize (B t x).
destruct B as [w B]; destruct B as [MAX U].
exists (valw w).
red; splits; try done.
* ins; splits; try done.
ins; erewrite U; edone.
* intro v'; ins; desf.
specialize (H w MAX); desf; lia.
Qed.

Lemma max_w_t_exists_unique_fun' G (WF: Wf G) t :
  exists f : Loc -> Event, forall x, w_max_t G t x (f x).
Proof.
  eapply choice.
  ins.
apply max_w_t_exists.
done.
Qed.

Lemma w_max_t_add_read_same_loc G G' t n w x l
  (WF: Wf G)  (CONS: sra_consistent G)
  (READ: is_r_l l) (NRMW: ~ is_rmw_l l) 
  (LOC:  loc_l l = x)
  (ADD: add G G' t n l w) 
  (CO : ~ dom_rel (co G ⨾ (hb G)^? ⨾ ⦗is_tid t⦘) w) : 
  w_max_t G' t x w.
Proof.
unfold w_max_t.
unfold set_compl.
unfold "∩₁"; splits.
- eapply add_acts_set; eauto. 
  left. apply ADD.
- apply ADD.
- red in ADD; desc. congruence.
- left. cdes ADD; desc.
  exists (ThreadEvent t n l), (ThreadEvent t n l).
  splits; vauto.
  right. eapply add_hb; try edone.
  basic_solver 12.
- intros.
intro X.
eapply add_non_write_dom_co_hb in X; try edone.
2: type_solver.
unfolder in X; desf.
+ 
eapply CO; basic_solver 12.
+ 
eapply CO; basic_solver 12.
+ eapply co_irr; eauto.
+ red in CONS; desc.
eapply IRR. apply ct_begin.
exists z; splits; try basic_solver.
apply rt_step. basic_solver.
Qed.

End Wmax_t.
